#include <sys/types.h>
#include <sys/qnx_glob.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <process.h>
#include <sched.h>
#include <stdio.h>
#include <i86.h>
#include <sys/kernel.h>
#include <sys/pci.h>
#include <sys/name.h>

#include "msg_pci.h"

/* _CA_PCI_MSG server */
/* needs to be compiled -T1 and run as root, probably make it setuid */
/* also, have it register a name... so only one is needed on system. */

void main()
{
   struct pci_msg msg;
   pid_t pid;
   int done = 0;

   /* register so people can find me... */
   if ( -1 == qnx_name_attach( 0, WORKER_NAME ))
   {
     if (errno == EBUSY)
       exit(1); /* one already exists */
     else
       perror("_CA_PCI_Worker unable to attach name, continuing ");
       /* continue in order to let the calling program work */
   }

   while(!done)
   {
       //fprintf(stderr,"before receive\n");
       pid = Receive(0, &msg, sizeof(msg));
       //fprintf(stderr,"after receive\n");
       if (pid == -1 )
       {
         perror("CA_PCI server Receive");
         exit(1);
       }

       switch( msg.req )
       {
          case PCI_BIOS_PRESENT:
            //fprintf(stderr,"bp\n");
            msg.status = _CA_PCI_BIOS_Present( &msg.req_info.bp.lastbus,
                                              &msg.req_info.bp.version,
                                              &msg.req_info.bp.hardware );
            break;
          case PCI_FIND_DEVICE:
            //fprintf(stderr,"fd\n");
            msg.status = _CA_PCI_Find_Device( msg.req_info.fd.devid,
                                              msg.req_info.fd.venid,
                                              msg.req_info.fd.index,
                                              &msg.req_info.fd.busnum,
                                              &msg.req_info.fd.devfuncnum );
            break;
          case PCI_FIND_CLASS:
            //fprintf(stderr,"fc\n");
            msg.status = _CA_PCI_Find_Class( msg.req_info.fc.ClassCode,
                                             msg.req_info.fc.index,
                                             &msg.req_info.fc.busnum,
                                             &msg.req_info.fc.devfuncnum );
            break;
          case PCI_GENERATE_SPECIAL_CYCLE:
            //fprintf(stderr,"gsc\n");
            msg.status = _CA_PCI_Generate_Special_Cycle( 
                                             msg.req_info.gsc.Busnum,
                                             msg.req_info.gsc.SpecialCycleData);
            break;
          case PCI_READ_CONFIG_BYTE:
            //fprintf(stderr,"rcb\n");
            msg.status = _CA_PCI_Read_Config_Byte( msg.req_info.rcb.busnum,
                                                   msg.req_info.rcb.devfuncnum,
                                                   msg.req_info.rcb.offset,
                                                   msg.req_info.rcb.cnt,
                                                   &msg.req_info.rcb.bufptr );
            break;
          case PCI_READ_CONFIG_WORD:
            //fprintf(stderr,"rcw\n");
            msg.status = _CA_PCI_Read_Config_Word( msg.req_info.rcw.busnum,
                                                   msg.req_info.rcw.devfuncnum,
                                                   msg.req_info.rcw.offset,
                                                   msg.req_info.rcw.cnt,
                                                   &msg.req_info.rcw.bufptr );
            break;
          case PCI_READ_CONFIG_DWORD:
            //fprintf(stderr,"rcd\n");
            msg.status = _CA_PCI_Read_Config_DWord( msg.req_info.rcd.busnum,
                                                    msg.req_info.rcd.devfuncnum,
                                                    msg.req_info.rcd.offset,
                                                    msg.req_info.rcd.cnt,
                                                    &msg.req_info.rcd.bufptr );
            break;
          case PCI_WRITE_CONFIG_BYTE:
            //fprintf(stderr,"wcb\n");
            msg.status = _CA_PCI_Write_Config_Byte( msg.req_info.wcb.busnum,
                                                    msg.req_info.wcb.devfuncnum,
                                                    msg.req_info.wcb.offset,
                                                    msg.req_info.wcb.cnt,
                                                    &msg.req_info.wcb.bufptr );
            break;
          case PCI_WRITE_CONFIG_WORD:
            //fprintf(stderr,"wcw\n");
            msg.status = _CA_PCI_Write_Config_Word( msg.req_info.wcw.busnum,
                                                    msg.req_info.wcw.devfuncnum,
                                                    msg.req_info.wcw.offset,
                                                    msg.req_info.wcw.cnt,
                                                    &msg.req_info.wcw.bufptr );
            break;
          case PCI_WRITE_CONFIG_DWORD:
            //fprintf(stderr,"wcd\n");
            msg.status = _CA_PCI_Write_Config_DWord( msg.req_info.wcd.busnum,
                                                     msg.req_info.wcd.devfuncnum,
                                                     msg.req_info.wcd.offset,
                                                     msg.req_info.wcd.cnt,
                                                     &msg.req_info.wcd.bufptr );
            break;
          case PCI_GET_IRQ_ROUTING_OPTIONS: {
            /* this one is ugly... may need to send extra data! */
            int size;
            struct _pci_route_buffer rbuf;
            char *ptr;

            //fprintf(stderr,"gro\n");
            size = msg.req_info.gro.rbufptr.BufferSize;
            if (size) {
               ptr = malloc(size);
               rbuf.DataBufferOffset = FP_OFF(ptr);
               rbuf.DataBufferSeg = FP_SEG(ptr);
               rbuf.BufferSize = size;
            } else 
            {
              rbuf.DataBufferOffset = 0;
              rbuf.DataBufferSeg = 0;
              rbuf.BufferSize = size;
            }
            msg.status = _CA_PCI_Get_Routing_Options( &rbuf,
                                                 &msg.req_info.gro.IRQForPCI );
            msg.req_info.gro.rbufptr.BufferSize = rbuf.BufferSize;
            Reply( pid, &msg, sizeof( msg ) );
            if (size) 
            {
               Receive( pid, &msg, sizeof(int) );
               Reply( pid, ptr, size );
               free( ptr );
            }
            //fprintf(stderr,"gro 2\n");
            continue;  /* don't fall through to common Reply() */
            break; }
          case PCI_SET_IRQ:
            //fprintf(stderr,"shi\n");
            msg.status = _CA_PCI_Set_Hardware_Interrupt( msg.req_info.shi.Busnum,
                                                         msg.req_info.shi.devfuncnum,
                                                         msg.req_info.shi.IntPin,
                                                         msg.req_info.shi.IRQNum );
            break;
          case WORKER_EXIT:
            done = 1;
            break;
       }
       Reply( pid, &msg, sizeof( msg ) );
   }
   exit(0);
}
