#include <sys/types.h>
#include <sys/qnx_glob.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <process.h>
#include <sched.h>
#include <sys/kernel.h>

#include "msg_pci.h"

static pid_t worker_pid = 0;

/* worker needs to be in path and a setuid root program */
static int get_worker() 
{
   int status;
   int hold_flag;
   int pid;

   pid = qnx_name_locate(0, WORKER_NAME, 0, NULL );
   if (pid != -1)
   {
      worker_pid = pid;
      return 1;
   }

   hold_flag = qnx_spawn_options.flags;
   qnx_spawn_options.flags = _SPAWN_NOZOMBIE;

   status = spawnlp( P_NOWAIT, WORKER_PROC, WORKER_PROC, NULL );
   qnx_spawn_options.flags = hold_flag;
   if (status == -1 )
   {
      perror("unable to spawn CA_PCI_worker");
      return 0;
   }
   /* in case two come up "simultaneously", and ours losing the registration */
   pid = qnx_name_locate(0, WORKER_NAME, 0, NULL );
   if (pid != -1) 
      worker_pid = pid;
   else
      worker_pid = status;

   return 1;
}

static int find_worker()
{
   int status = 0;
   if( worker_pid )
   {
      if ( !kill( 0, worker_pid ) )
      { 
        status = 1;
      } else
      {
        if (errno != ESRCH )
        {
           status = 1;
        } else
        {
           status = get_worker();
        }
      }

   } else 
   {
     status = get_worker();
   }
   return status;
}

void pci_kill_worker()
{
    struct pci_msg msg;
    if( worker_pid )
    {
       msg.req = WORKER_EXIT;
       Send( worker_pid, &msg, &msg, sizeof(msg), sizeof(msg) );
    }
    worker_pid = 0;
}

int _CA_PCI_Find_Device(unsigned devid, unsigned venid, 
                        unsigned index, unsigned *busnum, 
                        unsigned *devfuncnum)
{
   struct pci_msg msg;

   if( find_worker() )
   {
      msg.req = PCI_FIND_DEVICE;
      msg.req_info.fd.devid = devid;
      msg.req_info.fd.venid = venid;
      msg.req_info.fd.index = index;
      if (busnum) msg.req_info.fd.busnum = *busnum;
      if (devfuncnum) msg.req_info.fd.devfuncnum = *devfuncnum;
      if (Send(worker_pid, &msg, &msg, sizeof(msg), sizeof(msg)) != -1)
      {
         if (busnum) *busnum = msg.req_info.fd.busnum;
         if( devfuncnum) *devfuncnum = msg.req_info.fd.devfuncnum;
         return msg.status;
      } 
   }
   return PCI_NO_WORKER;
}

int _CA_PCI_Read_Config_Byte(unsigned busnum, unsigned devfuncnum,
                             unsigned offset, unsigned cnt, char *bufptr)
{
   struct pci_msg msg;

   if( find_worker() )
   {
      msg.req = PCI_READ_CONFIG_BYTE;
      msg.req_info.rcb.busnum = busnum;
      msg.req_info.rcb.devfuncnum = devfuncnum;
      msg.req_info.rcb.offset = offset;
      msg.req_info.rcb.cnt = cnt;
      memcpy( &msg.req_info.rcb.bufptr, bufptr, cnt );
      if (Send(worker_pid, &msg, &msg, sizeof(msg), sizeof(msg)) != -1)
      {
         if (bufptr) memcpy( bufptr, &msg.req_info.rcb.bufptr, cnt ); 
         return msg.status;
      } 
   }
   return PCI_NO_WORKER;

}

int _CA_PCI_Read_Config_Word(unsigned busnum, unsigned devfuncnum,
                             unsigned offset, unsigned cnt, char *bufptr)
{
   struct pci_msg msg;

   if( find_worker() )
   {
      msg.req = PCI_READ_CONFIG_WORD;
      msg.req_info.rcw.busnum = busnum;
      msg.req_info.rcw.devfuncnum = devfuncnum;
      msg.req_info.rcw.offset = offset;
      msg.req_info.rcw.cnt = cnt;
      memcpy( &msg.req_info.rcw.bufptr, bufptr, cnt*2 );
      if (Send(worker_pid, &msg, &msg, sizeof(msg), sizeof(msg)) != -1)
      {
         if (bufptr) memcpy( bufptr, &msg.req_info.rcw.bufptr, cnt*2 ); 
         return msg.status;
      } 
   }
   return PCI_NO_WORKER;

}

int _CA_PCI_Read_Config_DWord(unsigned busnum, unsigned devfuncnum,
                              unsigned offset, unsigned cnt, char *bufptr)
{
   struct pci_msg msg;

   if( find_worker() )
   {
      msg.req = PCI_READ_CONFIG_DWORD;
      msg.req_info.rcd.busnum = busnum;
      msg.req_info.rcd.devfuncnum = devfuncnum;
      msg.req_info.rcd.offset = offset;
      msg.req_info.rcd.cnt = cnt;
      memcpy( &msg.req_info.rcd.bufptr, bufptr, cnt*4 );
      if (Send(worker_pid, &msg, &msg, sizeof(msg), sizeof(msg)) != -1)
      {
         if (bufptr) memcpy( bufptr, &msg.req_info.rcd.bufptr, cnt*4 ); 
         return msg.status;
      } 
   }
   return PCI_NO_WORKER;

}


int _CA_PCI_Write_Config_Byte(unsigned busnum, unsigned devfuncnum,
                              unsigned offset, unsigned cnt, char *bufptr)
{
   struct pci_msg msg;

   if( find_worker() )
   {
      msg.req = PCI_WRITE_CONFIG_BYTE;
      msg.req_info.wcb.busnum = busnum;
      msg.req_info.wcb.devfuncnum = devfuncnum;
      msg.req_info.wcb.offset = offset;
      msg.req_info.wcb.cnt = cnt;
      memcpy( &msg.req_info.wcb.bufptr, bufptr, cnt ); 
      if (Send(worker_pid, &msg, &msg, sizeof(msg), sizeof(msg)) != -1)
      {
         return msg.status;
      } 
   }
   return PCI_NO_WORKER;
}


int _CA_PCI_Write_Config_Word(unsigned busnum, unsigned devfuncnum,
                              unsigned offset, unsigned cnt, char *bufptr)
{
   struct pci_msg msg;

   if( find_worker() )
   {
      msg.req = PCI_WRITE_CONFIG_WORD;
      msg.req_info.wcw.busnum = busnum;
      msg.req_info.wcw.devfuncnum = devfuncnum;
      msg.req_info.wcw.offset = offset;
      msg.req_info.wcw.cnt = cnt;
      memcpy( &msg.req_info.wcw.bufptr, bufptr, cnt*2 ); 
      if (Send(worker_pid, &msg, &msg, sizeof(msg), sizeof(msg)) != -1)
      {
         return msg.status;
      } 
   }
   return PCI_NO_WORKER;
}

int _CA_PCI_Write_Config_DWord(unsigned busnum, unsigned devfuncnum,
                               unsigned offset, unsigned cnt, char *bufptr)
{
   struct pci_msg msg;

   if( find_worker() )
   {
      msg.req = PCI_WRITE_CONFIG_DWORD;
      msg.req_info.wcd.busnum = busnum;
      msg.req_info.wcd.devfuncnum = devfuncnum;
      msg.req_info.wcd.offset = offset;
      msg.req_info.wcd.cnt = cnt;
      memcpy( &msg.req_info.wcd.bufptr, bufptr, cnt*4 ); 
      if (Send(worker_pid, &msg, &msg, sizeof(msg), sizeof(msg)) != -1)
      {
         return msg.status;
      } 
   }
   return PCI_NO_WORKER;
}

int _CA_PCI_Generate_Special_Cycle(unsigned Busnum,
                                   unsigned long SpecialCycleData)
{
   struct pci_msg msg;

   if( find_worker() )
   {
      msg.req = PCI_GENERATE_SPECIAL_CYCLE;
      msg.req_info.gsc.Busnum = Busnum;
      msg.req_info.gsc.SpecialCycleData = SpecialCycleData;
      if (Send(worker_pid, &msg, &msg, sizeof(msg), sizeof(msg)) != -1)
      {
         return msg.status;
      } 
   }
   return PCI_NO_WORKER;
}

/* not documented, may crash some machines */
int _CA_PCI_Set_Hardware_Interrupt(unsigned Busnum, unsigned devfuncnum,
                                   unsigned IntPin, unsigned IRQNum)
{
   struct pci_msg msg;

   if( find_worker() )
   {
      msg.req = PCI_SET_IRQ;
      msg.req_info.shi.Busnum = Busnum;
      msg.req_info.shi.devfuncnum = devfuncnum;
      msg.req_info.shi.IntPin = IntPin;
      msg.req_info.shi.IRQNum = IRQNum;
      if (Send(worker_pid, &msg, &msg, sizeof(msg), sizeof(msg)) != -1)
      {
         return msg.status;
      } 
   }
   return PCI_NO_WORKER;
}

/* not documented, may crash some machines */
/* this one is ugly... route buffer is a struct containing a size and /*
/* far pointer ... need some smarts */
/* And that bloody far pointer isn't... it is 16bit offset -- can't work */
/* in our context.  Create _r version that works, and return failure for */
/* the normal one. */

int _CA_PCI_Get_Routing_Options(struct _pci_route_buffer *rbufptr,
                                unsigned *IRQForPCI)
{
   /* prevent unreferenced warnings */
   rbufptr = rbufptr;
   IRQForPCI = IRQForPCI;

   return PCI_UNSUPPORTED_FUNCT;
}

int _CA_PCI_Get_Routing_Options_r(struct _pci_route_buffer_r *rbufptr,
                                unsigned *IRQForPCI)
{
   struct pci_msg msg;

   if( find_worker() )
   {
      int size;

      size = rbufptr->BufferSize;
      msg.req = PCI_GET_IRQ_ROUTING_OPTIONS;
      msg.req_info.gro.rbufptr = *rbufptr;
      if (IRQForPCI) msg.req_info.gro.IRQForPCI = *IRQForPCI;
      if (Send(worker_pid, &msg, &msg, sizeof(msg), sizeof(msg)) != -1)
      {
         *rbufptr = msg.req_info.gro.rbufptr;
         if (IRQForPCI) *IRQForPCI = msg.req_info.gro.IRQForPCI;
         /* get the additional data, if any... */
         if (size) Send(worker_pid, &size, (void *)rbufptr->DataBufferOffset, 
                        sizeof(int), size );
         return msg.status;
      } 
   }
   return PCI_NO_WORKER;

}

int _CA_PCI_Find_Class(unsigned long ClassCode, unsigned index,
                       unsigned *busnum, unsigned *devfuncnum)
{
   struct pci_msg msg;

   if( find_worker() )
   {
      msg.req = PCI_FIND_CLASS;
      msg.req_info.fc.ClassCode = ClassCode;
      msg.req_info.fc.index = index;
      if (busnum) msg.req_info.fc.busnum = *busnum;
      if (devfuncnum) msg.req_info.fc.devfuncnum = *devfuncnum;
      if (Send(worker_pid, &msg, &msg, sizeof(msg), sizeof(msg)) != -1)
      {
         if (busnum) *busnum = msg.req_info.fc.busnum;
         if (devfuncnum) *devfuncnum = msg.req_info.fc.devfuncnum;
         return msg.status;
      } 
   }
   return PCI_NO_WORKER;
}


int _CA_PCI_BIOS_Present(unsigned *lastbus, unsigned *version,
                         unsigned *hardware)
{
   struct pci_msg msg;

   if( find_worker() )
   {
      msg.req = PCI_BIOS_PRESENT;
      if (lastbus) msg.req_info.bp.lastbus = *lastbus;
      if (version) msg.req_info.bp.version = *version;
      if (hardware) msg.req_info.bp.hardware = *hardware;
      if (Send(worker_pid, &msg, &msg, sizeof(msg), sizeof(msg)) != -1)
      {
         if (lastbus) *lastbus = msg.req_info.bp.lastbus;
         if (version) *version = msg.req_info.bp.version;
         if (hardware) *hardware = msg.req_info.bp.hardware;
         return msg.status;
      } 
   }
   return PCI_NO_WORKER;
}
